import os
import re
from datetime import timedelta, datetime
from typing import List, Any, Optional

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, rsa
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat, NoEncryption, load_pem_private_key
from cryptography.x509 import ExtendedKeyUsageOID, NameOID


EC_SUPPORTED = {}
EC_SUPPORTED.update([(curve.name.upper(), curve) for curve in [
    ec.SECP192R1,
    ec.SECP224R1,
    ec.SECP256R1,
    ec.SECP384R1,
]])


def _private_key(key_type):
    if isinstance(key_type, str):
        key_type = key_type.upper()
        m = re.match(r'^(RSA)?(\d+)$', key_type)
        if m:
            key_type = int(m.group(2))

    if isinstance(key_type, int):
        return rsa.generate_private_key(
            public_exponent=65537,
            key_size=key_type,
            backend=default_backend()
        )
    if not isinstance(key_type, ec.EllipticCurve) and key_type in EC_SUPPORTED:
        key_type = EC_SUPPORTED[key_type]
    return ec.generate_private_key(
        curve=key_type,
        backend=default_backend()
    )


class CertificateSpec:

    def __init__(self, name: str = None, domains: List[str] = None,
                 email: str = None,
                 key_type: str = None, single_file: bool = False,
                 valid_from: timedelta = timedelta(days=-1),
                 valid_to: timedelta = timedelta(days=89),
                 client: bool = False,
                 sub_specs: List['CertificateSpec'] = None):
        self._name = name
        self.domains = domains
        self.client = client
        self.email = email
        self.key_type = key_type
        self.single_file = single_file
        self.valid_from = valid_from
        self.valid_to = valid_to
        self.sub_specs = sub_specs

    @property
    def name(self) -> Optional[str]:
        if self._name:
            return self._name
        elif self.domains:
            return self.domains[0]
        return None

    @property
    def type(self) -> Optional[str]:
        if self.domains and len(self.domains):
            return "server"
        elif self.client:
            return "client"
        elif self.name:
            return "ca"
        return None


class Credentials:

    def __init__(self, name: str, cert: Any, pkey: Any, issuer: 'Credentials' = None):
        self._name = name
        self._cert = cert
        self._pkey = pkey
        self._issuer = issuer
        self._cert_file = None
        self._pkey_file = None
        self._store = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def subject(self) -> x509.Name:
        return self._cert.subject

    @property
    def key_type(self):
        if isinstance(self._pkey, RSAPrivateKey):
            return f"rsa{self._pkey.key_size}"
        elif isinstance(self._pkey, EllipticCurvePrivateKey):
            return f"{self._pkey.curve.name}"
        else:
            raise Exception(f"unknown key type: {self._pkey}")

    @property
    def private_key(self) -> Any:
        return self._pkey

    @property
    def certificate(self) -> Any:
        return self._cert

    @property
    def cert_pem(self) -> bytes:
        return self._cert.public_bytes(Encoding.PEM)

    @property
    def pkey_pem(self) -> bytes:
        return self._pkey.private_bytes(
            Encoding.PEM,
            PrivateFormat.TraditionalOpenSSL if self.key_type.startswith('rsa') else PrivateFormat.PKCS8,
            NoEncryption())

    @property
    def issuer(self) -> Optional['Credentials']:
        return self._issuer

    def set_store(self, store: 'CertStore'):
        self._store = store

    def set_files(self, cert_file: str, pkey_file: str = None):
        self._cert_file = cert_file
        self._pkey_file = pkey_file

    @property
    def cert_file(self) -> str:
        return self._cert_file

    @property
    def pkey_file(self) -> Optional[str]:
        return self._pkey_file

    def get_first(self, name) -> Optional['Credentials']:
        creds = self._store.get_credentials_for_name(name) if self._store else []
        return creds[0] if len(creds) else None

    def get_credentials_for_name(self, name) -> List['Credentials']:
        return self._store.get_credentials_for_name(name) if self._store else []

    def issue_certs(self, specs: List[CertificateSpec],
                    chain: List['Credentials'] = None) -> List['Credentials']:
        return [self.issue_cert(spec=spec, chain=chain) for spec in specs]

    def issue_cert(self, spec: CertificateSpec, chain: List['Credentials'] = None) -> 'Credentials':
        key_type = spec.key_type if spec.key_type else self.key_type
        creds = None
        if self._store:
            creds = self._store.load_credentials(
                name=spec.name, key_type=key_type, single_file=spec.single_file, issuer=self)
        if creds is None:
            creds = Ngtcp2TestCA.create_credentials(spec=spec, issuer=self, key_type=key_type,
                                                    valid_from=spec.valid_from, valid_to=spec.valid_to)
            if self._store:
                self._store.save(creds, single_file=spec.single_file)
                if spec.type == "ca":
                    self._store.save_chain(creds, "ca", with_root=True)

        if spec.sub_specs:
            if self._store:
                sub_store = CertStore(fpath=os.path.join(self._store.path, creds.name))
                creds.set_store(sub_store)
            subchain = chain.copy() if chain else []
            subchain.append(self)
            creds.issue_certs(spec.sub_specs, chain=subchain)
        return creds


class CertStore:

    def __init__(self, fpath: str):
        self._store_dir = fpath
        if not os.path.exists(self._store_dir):
            os.makedirs(self._store_dir)
        self._creds_by_name = {}

    @property
    def path(self) -> str:
        return self._store_dir

    def save(self, creds: Credentials, name: str = None,
             chain: List[Credentials] = None,
             single_file: bool = False) -> None:
        name = name if name is not None else creds.name
        cert_file = self.get_cert_file(name=name, key_type=creds.key_type)
        pkey_file = self.get_pkey_file(name=name, key_type=creds.key_type)
        if single_file:
            pkey_file = None
        with open(cert_file, "wb") as fd:
            fd.write(creds.cert_pem)
            if chain:
                for c in chain:
                    fd.write(c.cert_pem)
            if pkey_file is None:
                fd.write(creds.pkey_pem)
        if pkey_file is not None:
            with open(pkey_file, "wb") as fd:
                fd.write(creds.pkey_pem)
        creds.set_files(cert_file, pkey_file)
        self._add_credentials(name, creds)

    def save_chain(self, creds: Credentials, infix: str, with_root=False):
        name = creds.name
        chain = [creds]
        while creds.issuer is not None:
            creds = creds.issuer
            chain.append(creds)
        if not with_root and len(chain) > 1:
            chain = chain[:-1]
        chain_file = os.path.join(self._store_dir, f'{name}-{infix}.pem')
        with open(chain_file, "wb") as fd:
            for c in chain:
                fd.write(c.cert_pem)

    def _add_credentials(self, name: str, creds: Credentials):
        if name not in self._creds_by_name:
            self._creds_by_name[name] = []
        self._creds_by_name[name].append(creds)

    def get_credentials_for_name(self, name) -> List[Credentials]:
        return self._creds_by_name[name] if name in self._creds_by_name else []

    def get_cert_file(self, name: str, key_type=None) -> str:
        key_infix = ".{0}".format(key_type) if key_type is not None else ""
        return os.path.join(self._store_dir, f'{name}{key_infix}.cert.pem')

    def get_pkey_file(self, name: str, key_type=None) -> str:
        key_infix = ".{0}".format(key_type) if key_type is not None else ""
        return os.path.join(self._store_dir, f'{name}{key_infix}.pkey.pem')

    def load_pem_cert(self, fpath: str) -> x509.Certificate:
        with open(fpath) as fd:
            return x509.load_pem_x509_certificate("".join(fd.readlines()).encode())

    def load_pem_pkey(self, fpath: str):
        with open(fpath) as fd:
            return load_pem_private_key("".join(fd.readlines()).encode(), password=None)

    def load_credentials(self, name: str, key_type=None, single_file: bool = False, issuer: Credentials = None):
        cert_file = self.get_cert_file(name=name, key_type=key_type)
        pkey_file = cert_file if single_file else self.get_pkey_file(name=name, key_type=key_type)
        if os.path.isfile(cert_file) and os.path.isfile(pkey_file):
            cert = self.load_pem_cert(cert_file)
            pkey = self.load_pem_pkey(pkey_file)
            creds = Credentials(name=name, cert=cert, pkey=pkey, issuer=issuer)
            creds.set_store(self)
            creds.set_files(cert_file, pkey_file)
            self._add_credentials(name, creds)
            return creds
        return None


class Ngtcp2TestCA:

    @classmethod
    def create_root(cls, name: str, store_dir: str, key_type: str = "rsa2048") -> Credentials:
        store = CertStore(fpath=store_dir)
        creds = store.load_credentials(name="ca", key_type=key_type, issuer=None)
        if creds is None:
            creds = Ngtcp2TestCA._make_ca_credentials(name=name, key_type=key_type)
            store.save(creds, name="ca")
            creds.set_store(store)
        return creds

    @staticmethod
    def create_credentials(spec: CertificateSpec, issuer: Credentials, key_type: Any,
                           valid_from: timedelta = timedelta(days=-1),
                           valid_to: timedelta = timedelta(days=89),
                           ) -> Credentials:
        """Create a certificate signed by this CA for the given domains.
        :returns: the certificate and private key PEM file paths
        """
        if spec.domains and len(spec.domains):
            creds = Ngtcp2TestCA._make_server_credentials(name=spec.name, domains=spec.domains,
                                                          issuer=issuer, valid_from=valid_from,
                                                          valid_to=valid_to, key_type=key_type)
        elif spec.client:
            creds = Ngtcp2TestCA._make_client_credentials(name=spec.name, issuer=issuer,
                                                          email=spec.email, valid_from=valid_from,
                                                          valid_to=valid_to, key_type=key_type)
        elif spec.name:
            creds = Ngtcp2TestCA._make_ca_credentials(name=spec.name, issuer=issuer,
                                                      valid_from=valid_from, valid_to=valid_to,
                                                      key_type=key_type)
        else:
            raise Exception(f"unrecognized certificate specification: {spec}")
        return creds

    @staticmethod
    def _make_x509_name(org_name: str = None, common_name: str = None, parent: x509.Name = None) -> x509.Name:
        name_pieces = []
        if org_name:
            oid = NameOID.ORGANIZATIONAL_UNIT_NAME if parent else NameOID.ORGANIZATION_NAME
            name_pieces.append(x509.NameAttribute(oid, org_name))
        elif common_name:
            name_pieces.append(x509.NameAttribute(NameOID.COMMON_NAME, common_name))
        if parent:
            name_pieces.extend([rdn for rdn in parent])
        return x509.Name(name_pieces)

    @staticmethod
    def _make_csr(
            subject: x509.Name,
            pkey: Any,
            issuer_subject: Optional[Credentials],
            valid_from_delta: timedelta = None,
            valid_until_delta: timedelta = None
    ):
        pubkey = pkey.public_key()
        issuer_subject = issuer_subject if issuer_subject is not None else subject

        valid_from = datetime.now()
        if valid_until_delta is not None:
            valid_from += valid_from_delta
        valid_until = datetime.now()
        if valid_until_delta is not None:
            valid_until += valid_until_delta

        return (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer_subject)
            .public_key(pubkey)
            .not_valid_before(valid_from)
            .not_valid_after(valid_until)
            .serial_number(x509.random_serial_number())
            .add_extension(
                x509.SubjectKeyIdentifier.from_public_key(pubkey),
                critical=False,
            )
        )

    @staticmethod
    def _add_ca_usages(csr: Any) -> Any:
        return csr.add_extension(
            x509.BasicConstraints(ca=True, path_length=9),
            critical=True,
        ).add_extension(
            x509.KeyUsage(
                digital_signature=True,
                content_commitment=False,
                key_encipherment=False,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=True,
                crl_sign=True,
                encipher_only=False,
                decipher_only=False),
            critical=True
        ).add_extension(
            x509.ExtendedKeyUsage([
                ExtendedKeyUsageOID.CLIENT_AUTH,
                ExtendedKeyUsageOID.SERVER_AUTH,
                ExtendedKeyUsageOID.CODE_SIGNING,
            ]),
            critical=True
        )

    @staticmethod
    def _add_leaf_usages(csr: Any, domains: List[str], issuer: Credentials) -> Any:
        return csr.add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        ).add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(
                issuer.certificate.extensions.get_extension_for_class(
                    x509.SubjectKeyIdentifier).value),
            critical=False
        ).add_extension(
            x509.SubjectAlternativeName([x509.DNSName(domain) for domain in domains]),
            critical=True,
        ).add_extension(
            x509.ExtendedKeyUsage([
                ExtendedKeyUsageOID.SERVER_AUTH,
            ]),
            critical=True
        )

    @staticmethod
    def _add_client_usages(csr: Any, issuer: Credentials, rfc82name: str = None) -> Any:
        cert = csr.add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        ).add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(
                issuer.certificate.extensions.get_extension_for_class(
                    x509.SubjectKeyIdentifier).value),
            critical=False
        )
        if rfc82name:
            cert.add_extension(
                x509.SubjectAlternativeName([x509.RFC822Name(rfc82name)]),
                critical=True,
            )
        cert.add_extension(
            x509.ExtendedKeyUsage([
                ExtendedKeyUsageOID.CLIENT_AUTH,
            ]),
            critical=True
        )
        return cert

    @staticmethod
    def _make_ca_credentials(name, key_type: Any,
                             issuer: Credentials = None,
                             valid_from: timedelta = timedelta(days=-1),
                             valid_to: timedelta = timedelta(days=89),
                             ) -> Credentials:
        pkey = _private_key(key_type=key_type)
        if issuer is not None:
            issuer_subject = issuer.certificate.subject
            issuer_key = issuer.private_key
        else:
            issuer_subject = None
            issuer_key = pkey
        subject = Ngtcp2TestCA._make_x509_name(org_name=name, parent=issuer.subject if issuer else None)
        csr = Ngtcp2TestCA._make_csr(subject=subject,
                                     issuer_subject=issuer_subject, pkey=pkey,
                                     valid_from_delta=valid_from, valid_until_delta=valid_to)
        csr = Ngtcp2TestCA._add_ca_usages(csr)
        cert = csr.sign(private_key=issuer_key,
                        algorithm=hashes.SHA256(),
                        backend=default_backend())
        return Credentials(name=name, cert=cert, pkey=pkey, issuer=issuer)

    @staticmethod
    def _make_server_credentials(name: str, domains: List[str], issuer: Credentials,
                                 key_type: Any,
                                 valid_from: timedelta = timedelta(days=-1),
                                 valid_to: timedelta = timedelta(days=89),
                                 ) -> Credentials:
        name = name
        pkey = _private_key(key_type=key_type)
        subject = Ngtcp2TestCA._make_x509_name(common_name=name, parent=issuer.subject)
        csr = Ngtcp2TestCA._make_csr(subject=subject,
                                     issuer_subject=issuer.certificate.subject, pkey=pkey,
                                     valid_from_delta=valid_from, valid_until_delta=valid_to)
        csr = Ngtcp2TestCA._add_leaf_usages(csr, domains=domains, issuer=issuer)
        cert = csr.sign(private_key=issuer.private_key,
                        algorithm=hashes.SHA256(),
                        backend=default_backend())
        return Credentials(name=name, cert=cert, pkey=pkey, issuer=issuer)

    @staticmethod
    def _make_client_credentials(name: str,
                                 issuer: Credentials, email: Optional[str],
                                 key_type: Any,
                                 valid_from: timedelta = timedelta(days=-1),
                                 valid_to: timedelta = timedelta(days=89),
                                 ) -> Credentials:
        pkey = _private_key(key_type=key_type)
        subject = Ngtcp2TestCA._make_x509_name(common_name=name, parent=issuer.subject)
        csr = Ngtcp2TestCA._make_csr(subject=subject,
                                     issuer_subject=issuer.certificate.subject, pkey=pkey,
                                     valid_from_delta=valid_from, valid_until_delta=valid_to)
        csr = Ngtcp2TestCA._add_client_usages(csr, issuer=issuer, rfc82name=email)
        cert = csr.sign(private_key=issuer.private_key,
                        algorithm=hashes.SHA256(),
                        backend=default_backend())
        return Credentials(name=name, cert=cert, pkey=pkey, issuer=issuer)



import logging
import pytest

from .ngtcp2test import Env


@pytest.mark.usefixtures("env")
def pytest_report_header(config):
    env = Env()
    return [
        f"ngtcp2-examples: [{env.version}, crypto_libs={env.crypto_libs}]",
        f"example clients: {env.clients}",
        f"example servers: {env.servers}",
    ]


@pytest.fixture(scope="package")
def env(pytestconfig) -> Env:
    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    logging.getLogger('').addHandler(console)
    env = Env(pytestconfig=pytestconfig)
    level = logging.DEBUG if env.verbose > 0 else logging.INFO
    console.setLevel(level)
    logging.getLogger('').setLevel(level=level)
    env.setup()

    return env



import logging
import os
import re
import subprocess
import time
from datetime import datetime, timedelta
from threading import Thread

from .tls import HandShake
from .env import Env, CryptoLib
from .log import LogFile, HexDumpScanner


log = logging.getLogger(__name__)


class ServerRun:

    def __init__(self, env: Env, logfile: LogFile):
        self.env = env
        self._logfile = logfile
        self.log_lines = self._logfile.get_recent()
        self._data_recs = None
        self._hs_recs = None
        if self.env.verbose > 1:
            log.debug(f'read {len(self.log_lines)} lines from {logfile.path}')

    @property
    def handshake(self):
        if self._data_recs is None:
            self._data_recs = [data for data in HexDumpScanner(source=self.log_lines)]
            if self.env.verbose > 1:
                log.debug(f'detected {len(self._data_recs)} hexdumps '
                          f'in {self._logfile.path}')
        if self._hs_recs is None:
            self._hs_recs = [hrec for hrec in HandShake(source=self._data_recs,
                                                        verbose=self.env.verbose)]
            if self.env.verbose > 1:
                log.debug(f'detected {len(self._hs_recs)} crypto records '
                          f'in {self._logfile.path}')
        return self._hs_recs

    @property
    def hs_stripe(self):
        return ":".join([hrec.name for hrec in self.handshake])


def monitor_proc(env: Env, proc):
    _env = env
    proc.wait()


class ExampleServer:

    def __init__(self, env: Env, crypto_lib: str, verify_client=False):
        self.env = env
        self._crypto_lib = crypto_lib
        self._path = env.server_path(self._crypto_lib)
        self._logpath = f'{self.env.gen_dir}/{self._crypto_lib}-server.log'
        self._log = LogFile(path=self._logpath)
        self._logfile = None
        self._process = None
        self._verify_client = verify_client

    @property
    def path(self):
        return self._path

    @property
    def crypto_lib(self):
        return self._crypto_lib

    @property
    def uses_cipher_config(self):
        return CryptoLib.uses_cipher_config(self.crypto_lib)

    def supports_cipher(self, cipher):
        return CryptoLib.supports_cipher(self.crypto_lib, cipher)

    @property
    def log(self):
        return self._log

    def exists(self):
        return os.path.isfile(self.path)

    def start(self):
        if self._process is not None:
            return False
        creds = self.env.get_server_credentials()
        assert creds
        args = [
            self.path,
            f'--htdocs={self.env.htdocs_dir}',
        ]
        if self._verify_client:
            args.append('--verify-client')
        args.extend([
            '*', str(self.env.examples_port),
            creds.pkey_file, creds.cert_file
        ])
        self._logfile = open(self._logpath, 'w')
        self._process = subprocess.Popen(args=args, text=True,
                                         stdout=self._logfile, stderr=self._logfile)
        t = Thread(target=monitor_proc, daemon=True, args=(self.env, self._process))
        t.start()
        timeout = 5
        end = datetime.now() + timedelta(seconds=timeout)
        while True:
            if self._process.poll():
                return False
            try:
                if self.log.scan_recent(pattern=re.compile(r'^Using document root'), timeout=0.5):
                    break
            except TimeoutError:
                pass
            if datetime.now() > end:
                raise TimeoutError(f"pattern not found in error log after {timeout} seconds")
        self.log.advance()
        return True

    def stop(self):
        if self._process:
            self._process.terminate()
            self._process = None
        if self._logfile:
            self._logfile.close()
            self._logfile = None
        return True

    def restart(self):
        self.stop()
        self._log.reset()
        return self.start()

    def get_run(self) -> ServerRun:
        return ServerRun(env=self.env, logfile=self.log)



from .env import Env, CryptoLib
from .log import LogFile
from .client import ExampleClient, ClientRun
from .server import ExampleServer, ServerRun
from .certs import Ngtcp2TestCA, Credentials
from .tls import HandShake, HSRecord



import pytest

from .ngtcp2test import ExampleClient
from .ngtcp2test import ExampleServer
from .ngtcp2test import Env


@pytest.mark.skipif(condition=len(Env.get_crypto_libs()) == 0,
                    reason="no crypto lib examples configured")
class TestClientCert:

    @pytest.fixture(scope='class', params=Env.get_crypto_libs())
    def server(self, env, request) -> ExampleServer:
        s = ExampleServer(env=env, crypto_lib=request.param,
                          verify_client=True)
        assert s.exists(), f'server not found: {s.path}'
        assert s.start()
        yield s
        s.stop()

    @pytest.fixture(scope='function', params=Env.get_crypto_libs())
    def client(self, env, request) -> ExampleClient:
        client = ExampleClient(env=env, crypto_lib=request.param)
        assert client.exists()
        yield client

    def test_04_01(self, env: Env, server, client):
        # run GET with a server requesting a cert, client has none to offer
        cr = client.http_get(server, url=f'https://{env.example_domain}/')
        assert cr.returncode == 0
        cr.assert_verify_null_handshake()
        creqs = [r for r in cr.handshake if r.hsid == 13]  # CertificateRequest
        assert len(creqs) == 1
        creq = creqs[0].to_json()
        certs = [r for r in cr.server.handshake if r.hsid == 11]  # Certificate
        assert len(certs) == 1
        crec = certs[0].to_json()
        assert len(crec['certificate_list']) == 0
        assert creq['context'] == crec['context']
        # TODO: check that GET had no answer

    def test_04_02(self, env: Env, server, client):
        # run GET with a server requesting a cert, client has cert to offer
        credentials = env.ca.get_first("clientsX")
        cr = client.http_get(server, url=f'https://{env.example_domain}/',
                             credentials=credentials)
        assert cr.returncode == 0
        cr.assert_verify_cert_handshake()
        creqs = [r for r in cr.handshake if r.hsid == 13]  # CertificateRequest
        assert len(creqs) == 1
        creq = creqs[0].to_json()
        certs = [r for r in cr.server.handshake if r.hsid == 11]  # Certificate
        assert len(certs) == 1
        crec = certs[0].to_json()
        assert len(crec['certificate_list']) == 1
        assert creq['context'] == crec['context']
        # TODO: check that GET indeed gave a response



import logging
import os
from configparser import ConfigParser, ExtendedInterpolation
from typing import Dict, Optional

from .certs import CertificateSpec, Ngtcp2TestCA, Credentials

log = logging.getLogger(__name__)


class CryptoLib:

    IGNORES_CIPHER_CONFIG = [
        'picotls', 'boringssl'
    ]
    UNSUPPORTED_CIPHERS = {
        'wolfssl': [
            'TLS_AES_128_CCM_SHA256',  # no plans to
        ],
        'picotls': [
            'TLS_AES_128_CCM_SHA256',  # no plans to
        ],
        'boringssl': [
            'TLS_AES_128_CCM_SHA256',  # no plans to
        ]
    }
    GNUTLS_CIPHERS = {
        'TLS_AES_128_GCM_SHA256': 'AES-128-GCM',
        'TLS_AES_256_GCM_SHA384': 'AES-256-GCM',
        'TLS_CHACHA20_POLY1305_SHA256': 'CHACHA20-POLY1305',
        'TLS_AES_128_CCM_SHA256': 'AES-128-CCM',
    }

    @classmethod
    def uses_cipher_config(cls, crypto_lib):
        return crypto_lib not in cls.IGNORES_CIPHER_CONFIG

    @classmethod
    def supports_cipher(cls, crypto_lib, cipher):
        return crypto_lib not in cls.UNSUPPORTED_CIPHERS or \
               cipher not in cls.UNSUPPORTED_CIPHERS[crypto_lib]

    @classmethod
    def adjust_ciphers(cls, crypto_lib, ciphers: str) -> str:
        if crypto_lib == 'gnutls':
            gciphers = "NORMAL:-VERS-ALL:+VERS-TLS1.3:-CIPHER-ALL"
            for cipher in ciphers.split(':'):
                gciphers += f':+{cls.GNUTLS_CIPHERS[cipher]}'
            return gciphers
        return ciphers


def init_config_from(conf_path):
    if os.path.isfile(conf_path):
        config = ConfigParser(interpolation=ExtendedInterpolation())
        config.read(conf_path)
        return config
    return None


TESTS_PATH = os.path.dirname(os.path.dirname(__file__))
EXAMPLES_PATH = os.path.dirname(TESTS_PATH)
DEF_CONFIG = init_config_from(os.path.join(TESTS_PATH, 'config.ini'))


class Env:

    @classmethod
    def get_crypto_libs(cls, configurable_ciphers=None):
        names = [name for name in DEF_CONFIG['examples']
                 if DEF_CONFIG['examples'][name] == 'yes']
        if configurable_ciphers is not None:
            names = [n for n in names if CryptoLib.uses_cipher_config(n)]
        return names

    def __init__(self, examples_dir=None, tests_dir=None, config=None,
                 pytestconfig=None):
        self._verbose = pytestconfig.option.verbose if pytestconfig is not None else 0
        self._examples_dir = examples_dir if examples_dir is not None else EXAMPLES_PATH
        self._tests_dir = examples_dir if tests_dir is not None else TESTS_PATH
        self._gen_dir = os.path.join(self._tests_dir, 'gen')
        self.config = config if config is not None else DEF_CONFIG
        self._version = self.config['ngtcp2']['version']
        self._crypto_libs = [name for name in self.config['examples']
                             if self.config['examples'][name] == 'yes']
        self._clients = [self.config['clients'][lib] for lib in self._crypto_libs
                         if lib in self.config['clients']]
        self._servers = [self.config['servers'][lib] for lib in self._crypto_libs
                         if lib in self.config['servers']]
        self._examples_pem = {
            'key': 'xxx',
            'cert': 'xxx',
        }
        self._htdocs_dir = os.path.join(self._gen_dir, 'htdocs')
        self._tld = 'tests.ngtcp2.nghttp2.org'
        self._example_domain = f"one.{self._tld}"
        self._ca = None
        self._cert_specs = [
            CertificateSpec(domains=[self._example_domain], key_type='rsa2048'),
            CertificateSpec(name="clientsX", sub_specs=[
               CertificateSpec(name="user1", client=True),
            ]),
        ]

    def issue_certs(self):
        if self._ca is None:
            self._ca = Ngtcp2TestCA.create_root(name=self._tld,
                                               store_dir=os.path.join(self.gen_dir, 'ca'),
                                               key_type="rsa2048")
        self._ca.issue_certs(self._cert_specs)

    def setup(self):
        os.makedirs(self._gen_dir, exist_ok=True)
        os.makedirs(self._htdocs_dir, exist_ok=True)
        self.issue_certs()

    def get_server_credentials(self) -> Optional[Credentials]:
        creds = self.ca.get_credentials_for_name(self._example_domain)
        if len(creds) > 0:
            return creds[0]
        return None

    @property
    def verbose(self) -> int:
        return self._verbose

    @property
    def version(self) -> str:
        return self._version

    @property
    def gen_dir(self) -> str:
        return self._gen_dir

    @property
    def ca(self):
        return self._ca

    @property
    def htdocs_dir(self) -> str:
        return self._htdocs_dir

    @property
    def example_domain(self) -> str:
        return self._example_domain

    @property
    def examples_dir(self) -> str:
        return self._examples_dir

    @property
    def examples_port(self) -> int:
        return int(self.config['examples']['port'])

    @property
    def examples_pem(self) -> Dict[str, str]:
        return self._examples_pem

    @property
    def crypto_libs(self):
        return self._crypto_libs

    @property
    def clients(self):
        return self._clients

    @property
    def servers(self):
        return self._servers

    def client_name(self, crypto_lib):
        if crypto_lib in self.config['clients']:
            return self.config['clients'][crypto_lib]
        return None

    def client_path(self, crypto_lib):
        cname = self.client_name(crypto_lib)
        if cname is not None:
            return os.path.join(self.examples_dir, cname)
        return None

    def server_name(self, crypto_lib):
        if crypto_lib in self.config['servers']:
            return self.config['servers'][crypto_lib]
        return None

    def server_path(self, crypto_lib):
        sname = self.server_name(crypto_lib)
        if sname is not None:
            return os.path.join(self.examples_dir, sname)
        return None



import pytest

from .ngtcp2test import ExampleClient
from .ngtcp2test import ExampleServer
from .ngtcp2test import Env


@pytest.mark.skipif(condition=len(Env.get_crypto_libs()) == 0,
                    reason="no crypto lib examples configured")
class TestHandshake:

    @pytest.fixture(scope='class', params=Env.get_crypto_libs())
    def server(self, env, request) -> ExampleServer:
        s = ExampleServer(env=env, crypto_lib=request.param)
        assert s.exists(), f'server not found: {s.path}'
        assert s.start()
        yield s
        s.stop()

    @pytest.fixture(scope='function', params=Env.get_crypto_libs())
    def client(self, env, request) -> ExampleClient:
        client = ExampleClient(env=env, crypto_lib=request.param)
        assert client.exists()
        yield client

    def test_01_01_get(self, env: Env, server, client):
        # run simple GET, no sessions, needs to give full handshake
        cr = client.http_get(server, url=f'https://{env.example_domain}/')
        assert cr.returncode == 0
        cr.assert_non_resume_handshake()
